#!/usr/local/mascot/perl64/bin/perl
##############################################################################
# Web client interface to ms-searchcontrol.exe and more                      #
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright 1998-2022 Matrix Science Limited  All Rights Reserved.           #
#                                                                            #
# This program may be used and modified within the licensee's organisation   #
# provided that this copyright notice remains intact. Distribution of this   #
# program or parts thereof outside the licensee's organisation without the   #
# prior written consent of Matrix Science Limited is expressly forbidden.    #
##############################################################################
#    $Archive:: /www/cgi/client.pl                                         $ #
#     $Author: villek@matrixscience.com $ #
#       $Date: 2024-08-22 13:41:05 +0100 $ #
#   $Revision: 82732c08c283536fd6c271e3980ed9f0be381dab | MASCOT_REL_3_0_0-2024-09-25-0-gaf416f51a3 $ #
# $NoKeywords::                                                            $ #
##############################################################################
 use v5.14;
 use strict;                                                                #
##############################################################################
#
# All actions take an optional argument for Mascot security:
#  sessionID=s 
#
# To obtain a task ID for search submission requires one argument:
#  create_task_id
#
# A search status query requires two arguments:
#  status task_id=n
#
# A search stage query requires two arguments:
#  status_current_stage task_id=n
#
# To find the last error string if the task ended with an error:
#  last_error task_id=n
#
# To kill a search job requires two arguments:
#  kill_job task_id=n
#
# To obtain a result file name requires two arguments:
#  result_file_name task_id=n
#
# To obtain summary results for a completed search requires two arguments:
#  results task_id=n sessionID=s
#
# To obtain summary results without query information:
#  results_header task_id=n sessionID=s
#
# If PERCOLATE=1 is set as a search parameter, then 'results' and
# 'results_header' implicitly call refine_results. This may cause a
# long delay before any results are printed. To avoid the delay and
# get access to the refining progress, call refine_results, then poll
# refine_results_status until refining is complete.
#
# To obtain xml (Distiller 2.0) results for a completed search requires three arguments:
#  xmlresults task_id=n reporttop=[FILE|AUTO|num_hits]
#
# To obtain the version number of ms-searchcontrol.exe requires one argument:
#  version
#
# To set the priority of a search requires three arguments:
#  nice_job task_id=n nice_value=n
# The valid range for nice_value is -20 to 20
#
# To obtain raw results in MIME format requires two arguments:
#  result_file_mime task_id=n sessionID=s
#
# To obtain compressed raw results without ions data
# (MIME format) requires two arguments:
#  result_file_mime2 task_id=n sessionID=s
#
# To obtain raw results in SQL format requires two arguments:
#  result_file_sql task_id=n sessionID=s
#
# To obtain compressed raw results format without ions data
# (SQL format) requires two arguments:
#  result_file_sql2 task_id=n sessionID=s
#
# To obtain raw results in "Windows INI" format requires two arguments:
#  result_file_ini task_id=n
#
# To obtain the Percolator output files (*.pop) in MIME format:
#  result_file_pop task_id=n sessionID=s
#  result_file_pop file=f sessionID=s
# Where f is relative or absolute path to result file
#
# If PERCOLATE=1 is set as a search parameter, then result_file_pop
# implicitly calls refine_results. This may cause a long delay before
# anything is printed. To avoid the delay and get access to the
# refining progress, call refine_results, then poll
# refine_results_status until refining is complete.
#
# To obtain query data for a range of queries in msr:queries XML format 
# has two formats and requires three arguments:
#  querydata task_id=n query_number=a[[-b][,c[-d]]...]
#  querydata file=s query_number=a[[-b][,c[-d]]...]
# Where s is relative or absolute path to result file
#
# To obtain list of user parameters for a completed search (Daemon 2.2) requires two arguments:
#  user_params task_id=n
#
# To obtain the version numbers of key Mascot binaries as INI format list requires one argument:
#  version_all
#
# To obtain quantitation.xml from ms-status.exe requires one argument:
#  quant_xml
#
# To obtain pretty-printed quantitation method summary from ms-config.exe requires two arguments:
#  quant_method [method=s|file=s]
#
# To obtain pretty-printed modification definition from ms-config.exe requires three arguments:
#  modification mod_name=s file=s
#
# To obtain crosslinking.xml from ms-status.exe requires one argument:
#  crosslinking_xml
#
# To obtain pretty-printed crosslinking method summary from ms-config.exe requires two arguments:
#  crosslinking_method [method=s|file=s]
#
# To obtain pretty-printed linker definition from ms-config.exe requires three arguments:
#  linker mod_name=s file=s
#
# To obtain name of operating system, as returned by Perl, requires one argument
#  osname
#
# To discover whether a particular database is active requires two arguments
#  db_active db_name=s
# Returns true if active (though not necessarily 'In use'), nothing otherwise 
#
# To create a results collection set (to be used in a merged report):
#  fileset_create
#
# To append files to a results collection set:
#  fileset_append fileset_id=ID file=s file=s ...
#
# To remove a particular file from a results collection set:
#  fileset_remove fileset_id=ID file=s
#
# To retrieve all files in the results collection set (one filename per line):
#  fileset_list fileset_id=ID
# If the fileset contains no files, or if the fileset doesn't exist, this
# will return an empty document.
#
# To download a previously uploaded Distiller XML file (the 'filename'
# parameter must be a filename associated with the task ID):
#  download_distiller_xml task_id=n sessionID=s distiller_xml_filename=filename
#
# To upload a Distiller XML file related to a particular results file,
# make a POST request with the form parameters "task_id", "sessionID" and
# "distiller_xml" (the actual XML data). Optional parameters are
# "computer_name" and "original_directory", which are not used in any way, 
# except added to the log file. Both URL encoded input and multipart/form-data
# are supported; for example,
#
#   task_id=12345&sessionID=67890&distiller_xml=<URL encoded XML data>
#
# and
#
#   --AaB03x
#   Content-Disposition: form-data; name="task_id"
#
#   12345
#   --AaB03x
#   Content-Disposition: form-data; name="sessionID"
#
#   67890
#   --AaB03x
#   Content-Disposition: form-data; name="distiller_xml"; filename="foo.xml"
#   Content-Type: text/xml
#
#   (contents of the XML file)
#   --AaB03x--
#
# are both accepted, although it generally makes sense to use the multipart
# form to avoid problems with line endings and binary data.
#
# To retrieve log messages about previously uploaded Distiller XML files:
#  recent_distiller_xml task_id=n sessionID=s from=timestamp
#
# To check if an XML file has been uploaded for a task ID (you can give any
# number of task_id parameters):
#  distiller_xml_exists sessionID=s task_id=n
#
# To refine results with machine learning before loading any of the
# reports or creating cache files:
#  refine_results task_id=n
#
# To find whether refining is active:
#  refine_results_status task_id=n
#
# The refine_results command returns immediately with either a
# "request-accepted" message or an error string. If refining has already been
# triggered, no action is taken.
#
# The refine_results_status command returns one of the following messages:
#   - missing_results_filename (if the task ID has no results file name)
#   - not_created (if Percolator pop files don't exist)
#   - created (if Percolator pop files exist)
#   - creating=xx% (percentage complete if refining)
#   - error=[error_text] (if a refining error occurred -- this may be a multi-line string contaning more details)
#
# To create cache files for a new results file before loading any of the
# reports:
#  create_results_cache task_id=n
#
# To find whether a cache file is being created:
#  results_cache_status task_id=n
#
# The create_results_cache command returns with either a
# "request-accepted" message or an error string. If caching has already been
# triggered, no action is taken.
#
# Note that if PERCOLATE=1 is set as a search parameter, then create_results_cache
# calls refine_results implicitly. This may cause a long delay before the
# request-accepted message is printed. To avoid the delay and get access to
# the refining progress, call refine_results, then poll refine_results_status
# until refining is complete. After that, call create_results_cache.
#
# The results_cache_status command returns one of the following messages:
#   - not_created (if cache files don't exist)
#   - created (if cache files exist)
#   - creating=xx% (percentage complete if caching)
#   - caching_disabled (if caching is disabled for ms-searchcontrol.exe in mascot.dat)
#   - missing_resultsfile=[filename] (if the results file linked to the task ID does not exist)
#   - missing_results_filename (if the task ID has no results file name)
#   - error=nnn:[error_text] (if a caching error occurred -- nnn is the Mascot Parser error number)
#   - unknown_id (if task ID is invalid)
#   - searchcontrol-error=nnn (if ms-searchcontrol.exe returns an error -- see the status command above)
#

use constant {
  MAX_FILE_UPLOAD => 2**31, # in bytes
  FILESET_TIMEOUT => 86400 * 7, # one week

  DISTILLER_XML_LOGFILE => '../logs/distiller_xml_uploads.log',
  DISTILLER_MSGHEADER => [
    'Timestamp', 'Datetime', 'Success?', 'Task ID', 'Filepath', 
    'Computer name', 'Original directory', 'Message'
  ],
  DISTILLER_MSGFMT => "%d\t%s\t%1s\t%s\t%s\t%s\t%s\t%s\n",
  # Note: Don't change the format unless you also change how the log file is
  # parsed in recent_distiller_xml and distiller_xml_exists.
};

use msparser;
use Util::IPC qw(fork_child);
use CGI;
use Util::FileOps qw(perl_path);

BEGIN {
  # Setting CGI variables must come after importing the module.

  $CGI::POST_MAX = MAX_FILE_UPLOAD;

# set current directory to script location for FastTrack on NT
# chdir also handles a Windows drive letter change
  if ($ENV{'WINDIR'} && $0 =~ /(.*)[\\\/]/) {
    chdir($1);
  }

  my $datfile = msparser::ms_datfile->new();
  my $opts = $datfile->getMascotOptions();
  my $tmpdir = join('/', $opts->getInterFileRelPath(), 'cache');

  if (not -d $tmpdir) {
    mkdir $tmpdir or warn "Warning: Cannot create temporary directory '$tmpdir': $! ($^E)";
  }

  if (-d $tmpdir) {
    if (-w _) {
      no warnings 'once';
      $CGI::TMPDIRECTORY = $tmpdir;
    } else {
      warn "Warning: Temporary directory '$tmpdir' is not writable; using system temp for file uploads";
    }
  }

  $| = 1; # set autoflush on STDOUT
}

# required Perl packages
use Fcntl qw(:DEFAULT :flock :mode);
use File::stat;
use Digest::MD5;

my $thisScript = eval { new CGI };

if ($@) {
  print "Content-Type: text/plain\n\n";
  # E.g. "Malformed multipart POST"
  print "problem=$@";
  exit 1;
}

if ($thisScript->request_method eq 'POST') {
# pull in common subroutines
  require "./common_subs.pl";

# www.pl contains the code to decrypt result file names
# its existence indicates that this script is running on the public web site
  if (-e "./www.pl") {
    do "./www.pl";
  }

  if (defined(my $upload_filename = $thisScript->param('distiller_xml'))) {
    my %supported_types = map { $_ => 1 } qw(
      text/xml
      application/xml
    );

    my ($taskID, $sessionID, $session, $sessionID_parameter) = _retrieve_common_parameters($thisScript);

    my $computer_name = $thisScript->param('computer_name');
    my $original_directory = $thisScript->param('original_directory');

    if (not $session->isPermitted($msparser::ms_security_tasks::SECTASK_DISTILLERCLIENT)) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, '(unknown)', $computer_name, $original_directory,
          'Not permitted by Mascot Security'
      );
      print "problem=Not permitted by Mascot Security\n";
      exit 1;
    }

    if (not $session->isPermitted($msparser::ms_security_tasks::SECTASK_UPLOAD_DISTILLER_XML)) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, '(unknown)', $computer_name, $original_directory,
          'Not permitted by Mascot Security'
      );
      print "problem=Not permitted by Mascot Security\n";
      exit 1;
    }

    if ($taskID < 1) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, '(unknown)', $computer_name, $original_directory,
          "Invalid task ID ($taskID)"
      );
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

    my $resfilepath = '';

  # Call ms-searchcontrol.exe to get result file name.
  # ms-searchcontrol.exe will validate the session as well.
    if (open(my $pipe, '"../bin/ms-searchcontrol.exe" --result_file_name --taskID '.common_subs::shellQuote($taskID).' '.$sessionID_parameter.' |')) {
      while (<$pipe>) {
        if (/^filename=(.+)/i) {
          chomp($resfilepath = $1);
          last;
        }
      }
      1 while (<$pipe>); # flush the pipe so that zombies won't arise
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, '(unknown)', $computer_name, $original_directory,
          "Could not open socket to ms-searchcontrol.exe: $! ($^E)"
      );
      print "problem=Could not open socket to ms-searchcontrol.exe: $! ($^E)\n";
      exit 1;
    }

    if ($resfilepath eq '') {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, '(unknown)', $computer_name, $original_directory,
          'No result file name'
      );
      print "problem=no result file name\n";
      exit 1;
    }

    if (not -e $resfilepath) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, '(unknown)', $computer_name, $original_directory,
          "Result file does not exist ($resfilepath)"
      );
      print "problem=result file does not exist\n";
      exit 1;
    }

    my $info = $thisScript->uploadInfo($upload_filename);

    if (not defined $$info{'Content-Type'}
            or not $supported_types{ lc $$info{'Content-Type'} }) {
      my $type = $$info{'Content-Type'} || '';
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, '(unknown)', $computer_name, $original_directory,
          "Unknown content type ($type)"
      );
      print "problem=Unknown content type for distiller_xml\n";
      exit 1;
    }

    my $pseudofh = $thisScript->upload('distiller_xml');

    if (not defined $pseudofh) {
      if (my $status = $thisScript->cgi_error) {
        # e.g. "400 Bad Request (malformed multipart POST)"
        print $thisScript->header(-status => $status);
        _log_distiller_upload_failure(
            $taskID, '(unknown)', $computer_name, $original_directory, $status
        );
      } else {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        _log_distiller_upload_failure(
            $taskID, '(unknown)', $computer_name, $original_directory,
            'Invalid filehandle from CGI.pm'
        );
        print "problem=Invalid filehandle for distiller_xml\n";
      }
      exit 1;
    }

    my $inputfh = do {
      if ($pseudofh->can('handle')) {
        $pseudofh->handle
      } else { # old CGI.pm...
        IO::Handle->new_from_fd(fileno($pseudofh), "r")
      }
    };
    binmode($inputfh);

    my $file_md5 = Digest::MD5->new();
    eval { $file_md5->addfile($inputfh) };

    if ($@) {
      my $str = ''.$@;
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, '(unknown)', $computer_name, $original_directory,
          "Error while reading from input file: $str"
      );
      print "problem=Could not read input file: $str\n";
      exit 1;
    }

    my $xmlfilepath = _resfilepath_to_distiller_xml_filepath($resfilepath, $file_md5->hexdigest);

    seek($inputfh, 0, 0) or do {
      my $errstr = "$! ($^E)";
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, $xmlfilepath, $computer_name, $original_directory,
          "Error while reading from input file: $errstr"
      );
      print "problem=Could not read input file: $errstr\n";
      exit 1;
    };

    if (-e $xmlfilepath) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, $xmlfilepath, $computer_name, $original_directory,
          'Distiller XML file already exists'
      );
      print "problem=Distiller XML file '$xmlfilepath' already exists\n";
      exit 1;
    }

    my ($outputfh, $errormsg, undef, $warn_errormsg) = _acquire_write_lock($xmlfilepath);

    if (not $outputfh) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      _log_distiller_upload_failure(
          $taskID, $xmlfilepath, $computer_name, $original_directory,
          "Could not open '$xmlfilepath' for writing: $errormsg"
      );
      print "problem=Could not open '$xmlfilepath' for writing: $errormsg\n";
      warn $warn_errormsg;
      exit 1;
    }

    binmode($outputfh);

    while (1) {
      my $buffer;
      my $n = eval { $inputfh->read($buffer, 1048576) };

      if ($@) {
        # E.g. "Malformed multipart POST"
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        _log_distiller_upload_failure(
            $taskID, $xmlfilepath, $computer_name, $original_directory,
            "$@",
        );
        print "problem=$@\n";
        exit 1;
      }

      if (not defined($n)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        _log_distiller_upload_failure(
            $taskID, $xmlfilepath, $computer_name, $original_directory,
            "Read error: $! ($^E)"
        );
        print "problem=Read error: $! ($^E)\n";
        exit 1;
      }

      if ($n == 0) {
        last;
      }

      print $outputfh $buffer or do {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        _log_distiller_upload_failure(
            $taskID, $xmlfilepath, $computer_name, $original_directory,
            "Write error: $! ($^E)"
        );
        print "problem=Write error: $! ($^E)\n";
        exit 1;
      };
    }

    close($inputfh); # ignore possible error
    close($outputfh); # ignore possible error

    print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
    _log_distiller_upload_success(
        $taskID, $xmlfilepath, $computer_name, $original_directory, 
    );

    print "success=Distiller XML file saved as '$xmlfilepath'\n";

    exit 0;

  } else {
    print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
    print "problem=Unknown command in POST request\n";
    exit 1;
  }

  exit 0;

} elsif ($thisScript->request_method eq 'GET' or not defined $thisScript->request_method) { # not defined when run on command line

  if (defined($thisScript->param('keywords'))) {
    $thisScript->param(-name => scalar($thisScript->param('keywords')),
                       -value => 1);
  }

  unless (defined($thisScript->param('quant_method'))
    || defined($thisScript->param('quant_xml'))
    || defined($thisScript->param('modification'))
    || defined($thisScript->param('crosslinking_method'))
    || defined($thisScript->param('crosslinking_xml'))
    || defined($thisScript->param('linker'))
    || defined($thisScript->param('result_file_mime2'))
    || defined($thisScript->param('result_file_sql2'))
    || defined($thisScript->param('result_file_pop'))
    || defined($thisScript->param('download_distiller_xml')) 
    || defined($thisScript->param('refine_results'))
    || defined($thisScript->param('create_results_cache'))){
    print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
  }

# pull in common subroutines
  require "./common_subs.pl";

# www.pl contains the code to decrypt result file names
# its existence indicates that this script is running on the public web site
  if (-e "./www.pl") {
    do "./www.pl";
  }

  my ($taskID, $sessionID, $session, $sessionID_parameter) = _retrieve_common_parameters($thisScript);

# branch according to action keyword      
  if (defined($thisScript->param('create_task_id'))) {

  # supply the correct action string for the search
    do {
      print "actionstring=nph-mascot.exe\n";
    };
  
  # Get a task ID
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --create_task_id $sessionID_parameter |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    print @temp;

  } elsif (defined($thisScript->param('status'))) {
    
    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

  # Call ms-searchcontrol.exe to get status info
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --status --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    print @temp;
  } elsif (defined($thisScript->param('status_current_stage'))) {
    
    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

  # Call ms-searchcontrol.exe to get search stage info (bug 11864)
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --status_current_stage --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    print @temp;
  } elsif (defined($thisScript->param('refine_results_status'))) {
    
    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

    my $resfilepath = fetch_result_file_name($taskID);

    if (not defined $resfilepath) {
        print "missing_results_filename\n";
        exit 0;
    }

    my $status = refiner_status($resfilepath);
    print $status;
    exit 0;

  } elsif (defined($thisScript->param('refine_results'))) {

    if ($taskID < 1) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Invalid task ID '$taskID'\n";
        exit 1;
    }

    my $resfilepath = fetch_result_file_name($taskID);

    if (not defined $resfilepath) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "missing_results_filename\n";
        exit 1;
    }

    my $waitpid = eval { fork_child(perl_path(), '../bin/refine_results_with_ml.pl', '--refine', '--resfile', $resfilepath) };

    if ($@) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=$@\n";
        exit 1;
    }

    my $string = "request-accepted\n";
    my $len = do {
      use bytes;
      length($string);
    };

    print $thisScript->header( -type => "text/plain", -charset => "utf-8", -content_length => $len );

    # Remove the flag file now if it exists. We assume triggering
    # the refining process means no e-mail needs to be sent. 
    #
    # Note that it's very important that this action is done before the "print
    # $string" line. See create_results_cache command.
    remove_flag_file_if_exists($resfilepath);

    # Fixed-length output must be printed now, so that the browser is
    # satisfied and waits for no more input.
    print $string;

    # And now we wait until the child process is done.
    $waitpid->();
    exit 0;

  } elsif (defined($thisScript->param('results_cache_status'))) {
    
    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

    my $resfilepath = fetch_result_file_name($taskID);

    # Bug 14744: We must check whether refining is finished. If it's ongoing,
    # just reply with 'creating'.
    if (defined $resfilepath) {
        my $status = refiner_status($resfilepath);

        if ($status =~ /^created/) {
            # OK
        } elsif ($status =~ /^creating/) {
            print 'creating=0%';
            exit 0;
        } elsif ($status =~ /^not_created/) {
            print $status;
            exit 0;
        } else {
            # Must be an error.
            print $status;
            exit 0;
        }
    }

  # Call ms-searchcontrol.exe to get status info
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --results_cache_status --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    print @temp;
  } elsif (defined($thisScript->param('create_results_cache'))) {

    if ($taskID < 1) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Invalid task ID '$taskID'\n";
        exit 1;
    }

    my $resfilepath = fetch_result_file_name($taskID);

    if (defined $resfilepath) {
        # Bug 14744: Refine the results before calling ms-searchcontrol.exe.
        my $status = refine_it($resfilepath);

        if ($status =~ /^created/) {
            # OK
        } else {
            print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
            print $status;
            exit 1;
        }
    }

    my @sessionID;
    
    if ($session->isValid) {
      @sessionID = ("--sessionID", $sessionID);
    }

    my $waitpid = eval { fork_child("../bin/ms-searchcontrol.exe", "--create_results_cache", "--taskID", $taskID, @sessionID) };

    if ($@) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=$@\n";
        exit 1;
    }

    my $string = "request-accepted\n";
    my $len = do {
        use bytes;
        length($string);
    };

    print $thisScript->header( -type => "text/plain", -charset => "utf-8", -content_length => $len );

    # Bug 12266: remove the flag file now if it exists. We assume triggering
    # the caching process means no e-mail needs to be sent. 
    #
    # Note that it's very important that this action is done before the "print
    # $string" line. If you print the full number of bytes before the flag file
    # is deleted, IIS will immediately make the process exit before waitpid can
    # be called. On Windows 7 at least (and likely any newer version), this
    # leads to zombie processes that never reach their main() function.
    # So, you can do just about anything between fork_child() and "print
    # $string", but you must not do anything but waitpid after "print $string".
    remove_flag_file_if_exists($resfilepath);

    # Fixed-length output must be printed now, so that the browser is
    # satisfied and waits for no more input.
    print $string;

    # And now we wait until the child process is done.
    $waitpid->();
    exit 0;
  } elsif (defined($thisScript->param('last_error'))) {
    
    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

    my $resfilepath = '';

    do {
        my $taskID_str = common_subs::shellQuote($taskID);
        my $status = qx{../bin/ms-searchcontrol.exe --status --taskID $taskID_str $sessionID_parameter};

        if ($status =~ /error=(-?\d+)/) {
            # OK; job ended in an error.
        } else {
            print "\n"; # correct output is blank when no error occurred
            exit 1;
        }
    };

  # Call ms-searchcontrol.exe to get result file name.
  # ms-searchcontrol.exe will validate the session as well.
    if (open(my $pipe, '"../bin/ms-searchcontrol.exe" --result_file_name --taskID '.common_subs::shellQuote($taskID).' '.$sessionID_parameter.' |')) {
      while (<$pipe>) {
        if (/^filename=(.+)/i) {
          chomp($resfilepath = $1);
          last;
        }
      }
      1 while (<$pipe>); # flush the pipe so that zombies won't arise
    } else {
      print "problem=Could not open socket to ms-searchcontrol.exe: $! ($^E)\n";
      exit 1;
    }

    if ($resfilepath eq '') {
      print "problem=no result file name\n";
      exit 1;
    }

    my $logfilepath = $resfilepath.'.log';

    if (not -e $logfilepath) {
      print "\n"; # no error = empty line
      exit 0;
    }

    if (open(my $fh, '<', $logfilepath)) {
      my $line;
      while (<$fh>) {
        # discard all but last line
        $line = $_;
      }
      print $line;
    } else {
      print "problem=Could not open batch log file: $! ($^E)\n";
      exit 1;
    }

  } elsif (defined($thisScript->param('kill_job'))) {

  # security is handled by ms-searchcontrol.exe

    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

  # Call ms-searchcontrol.exe to kill a job
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --kill_job --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    print @temp;

  } elsif (defined($thisScript->param('result_file_name'))) {
    
    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

  # Call ms-searchcontrol.exe to get result file name
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --result_file_name --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    print @temp;

  } elsif (defined($thisScript->param('results')) 
        || defined($thisScript->param('results_header'))
        || defined($thisScript->param('xmlresults'))
        || defined($thisScript->param('result_file_mime'))
        || defined($thisScript->param('result_file_sql'))
        || defined($thisScript->param('result_file_ini'))) {

  # security is handled by ms-searchcontrol.exe
  
    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

  # Same code for all except that xmlresults has additional argument
    my ($cmnd, @reporttop);
    if (defined($thisScript->param('xmlresults'))) {
      $cmnd = "--xmlresults";
      if (!defined($thisScript->param('reporttop'))) {
        print "problem=Invalid reporttop\n";
        exit 1;
      } else {
        @reporttop = ("--reporttop" , $thisScript->param('reporttop'));
      }
    } elsif (defined($thisScript->param('result_file_mime'))) {
      $cmnd = "--result_file_mime";
    } elsif (defined($thisScript->param('result_file_sql'))) {
      $cmnd = "--result_file_sql";
    } elsif (defined($thisScript->param('result_file_ini'))) {
      $cmnd = "--result_file_ini";
    } elsif (defined($thisScript->param('results_header'))) {
      $cmnd = "--results_header";
    } else {
      $cmnd = "--results";
    }
    
  # May need to decompress the result file
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --result_file_name --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    my $fileIn = "";
    foreach (@temp) {
      if (/^filename=(.+)/i) {
        $fileIn = $1;
        last;
      }
    }
    if ($fileIn) {
      if (-e common_subs::decompress($fileIn)) {
      # Remove the .flg file so that results don't get emailed.
        unlink common_subs::decompress($fileIn) . ".flg";
      # Also remove the batch log file if it exists.
        unlink common_subs::decompress($fileIn) . ".log";
      } else {
        print "problem=result file does not exist\n";
        exit 1;
      }
    } else {
      print "problem=no result file name\n";
      exit 1;
    }
    
    if ($cmnd eq "--results_header" or $cmnd eq "--results") {
        # Bug 14744: Refine the results before calling ms-searchcontrol.exe.
        my $status = refine_it($fileIn);

        if ($status =~ /^created/) {
            # OK
        } else {
            print $status;
            exit 1;
        }
    }

  # Call ms-searchcontrol.exe to output search results to STDOUT
    my @cmd = ('../bin/ms-searchcontrol.exe', $cmnd, '--taskID', $taskID);
    if ($sessionID) {
      push @cmd, '--sessionID', $sessionID;
    }
    push @cmd, @reporttop;
    if (system(@cmd)) {
      print "problem=Error return from ms-searchcontrol.exe\n";
      exit 1;
    }
    
  } elsif (defined($thisScript->param('result_file_mime2'))
           or defined($thisScript->param('result_file_sql2'))) {

    if ($taskID < 1) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

    my ($ms_searchcontrol_cmd, $output_file_ext);

    if (defined($thisScript->param('result_file_mime2'))) {
        $ms_searchcontrol_cmd = '--result_file_mime';
        $output_file_ext = '.dat';
    } elsif (defined($thisScript->param('result_file_sql2'))) {
        $ms_searchcontrol_cmd = '--result_file_sql';
        $output_file_ext = '.sql';
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Unknown command\n";
      exit 1;
    }

  # May need to decompress the result file
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --result_file_name --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    my $fileIn = "";
    foreach (@temp) {
      if (/^filename=(.+)/i) {
        $fileIn = $1;
        last;
      }
    }

    my $decompressed_fileIn;
    if ($fileIn) {
      $decompressed_fileIn = common_subs::decompress($fileIn);
      if (-e $decompressed_fileIn) {
      # Remove the .flg file so that results don't get emailed.
        unlink $decompressed_fileIn . ".flg";
      # Also remove the batch log file if it exists.
        unlink $decompressed_fileIn . ".log";
      } else {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=result file does not exist\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=no result file name\n";
      exit 1;
    }

  # Read result file, compress using ms_zip, and write to STDOUT
  # Remove ions data from query blocks

    my $zipper = msparser::ms_zip->new(0);
    
    if (not $zipper->isValid) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=unable to initialise ms_zip: ", $zipper->getLastErrorString(), "\n";
      exit 1;
    }

    my $pipe_in;
    unless (open($pipe_in, "\"../bin/ms-searchcontrol.exe\" $ms_searchcontrol_cmd --exclude_ions_data --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }

    binmode($pipe_in);

    # Read the first few (say 10) lines. If there is an error, such as
    # permission fault, it should be the first or at most second line.
    my @initial_lines = read_10_lines($pipe_in);

    if (my $errnum = parse_ms_searchcontrol_error(\@initial_lines)) {
        my @desc = map_ms_searchcontrol_error_to_description($errnum);
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        say $_ for @desc;
        exit 1;
    }
        
    my ($newName) = $fileIn =~ /([^\\\/]+)(:?\.[^\.]+)$/;

    # Careful with file extension if this is SQL output. We are
    # printing SQL statements, not the raw bytes from the SQLite
    # database, so indicate it with file extension ".sql".
    $newName .= $output_file_ext;
    
    print $thisScript->header( 
      -type => "application/x-gzip",
      -attachment => "$newName.gz",
      -Content_Encoding => "identity"
    );

    binmode(STDOUT);

    # Feed in the first few lines, since they were not an error message.
    print $zipper->compressMore(join('', @initial_lines));
    @initial_lines = ();
    
    # The readline() method (<$pipe_in>) uses the PerlIO layer.
    # The read() method uses the same PerlIO layer, so it's OK to mix and match.
    #
    # Reading in chunks of MB is much more efficient than reading
    # line by line.
    
    while (1) {
      my $status = read($pipe_in, my $buffer, $msparser::ms_zip::MAX_UNCOMPRESSED_SIZE);
      last if not $status;
      print $zipper->compressMore($buffer);
      last if not $zipper->isValid();
    }
    
    while ($zipper->isValid()) {
      my $s = $zipper->compressMore('');
      last if $s eq '';
      print $s;
    }

  } elsif (defined($thisScript->param('result_file_pop'))) {
    
    my $fileIn;
    
    if (defined($thisScript->param('file'))) {
      $fileIn = $thisScript->param('file');
      $fileIn = undef unless $fileIn =~ m/^[^\x00-\x1f]+$/;
    } else {
      if ($taskID < 1) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Invalid task ID '$taskID'\n";
        exit 1;
      }
      unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --result_file_name --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Could not open socket to ms-searchcontrol.exe\n";
        exit 1;
      }
      my @temp = <SOCK>;
      close(SOCK);
      $fileIn = "";
      foreach (@temp) {
        if (/^filename=(.+)/i) {
          $fileIn = $1;
          last;
        }
      }
    }

    if ($fileIn) {
      unless (-e common_subs::decompress($fileIn)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=result file does not exist\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=no result file name\n";
      exit 1;
    }

    do {
        # Bug 14849: Refine the results here if they haven't been refined already.
        my $status = refine_it($fileIn);

        if ($status =~ /^created/) {
            # OK
        } else {
            print $status;
            exit 1;
        }
    };

  # create an configuration options object from mascot.dat
    my $objDatFile = new msparser::ms_datfile("../config/mascot.dat");
  # abort on fatal error. Note that isValid() is true if there are only warnings
    unless ($objDatFile->isValid) {
      my $errorList = "problem=Mascot Parser error(s) at 3:\n";
      my $objErr = $objDatFile->getErrorHandler();
      for (my $i = 1; $i <= $objErr->getNumberOfErrors(); $i++) {
        $errorList .= $objErr->getErrorNumber($i) 
          . ": " . common_subs::htmlEscape($objErr->getErrorString($i)) . "\n";
      }
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print $errorList;
      exit 1;
    }
    my $objMascotOptions = $objDatFile->getMascotOptions;

  # see whether to use ms_mascotresfilebase::RESFILE_USE_CACHE flag
    my $resfileFlags;
    if (common_subs::useResfileCache($objMascotOptions, 'client.pl')) {
      $resfileFlags = $msparser::ms_mascotresfilebase::RESFILE_USE_CACHE;
    } else {
      $resfileFlags = $msparser::ms_mascotresfilebase::RESFILE_NOFLAG;
    }
    my $resfileCacheDirectory = $objMascotOptions->getCacheDirectory;

  # Mascot security
    my $objResFile = _open_resfile(common_subs::decompress($fileIn), 
                                                $resfileFlags, 
                                                $resfileCacheDirectory);
    unless ($objResFile->isValid) {
      my $errorList = "problem=Mascot Parser error(s) at 4:\n";
      for (my $i = 1; $i <= $objResFile->getNumberOfErrors(); $i++) {
        $errorList .= $objResFile->getErrorNumber($i) 
          . ": " . common_subs::htmlEscape($objResFile->getErrorString($i)) . "\n";
      }
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print $errorList;
      exit 1;
    }
    my $objParams = new msparser::ms_searchparams($objResFile);
    if ($session->isValid) {
      unless ($session->canResultsFileBeViewed($objParams->getUSERID)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Not permitted by Mascot security\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Not logged in to Mascot security\n";
      exit 1;
    }
    
  # get the POP filename
    my $vecFiles = new msparser::VectorString;  # For std::vector<std::string>;    
    $objResFile->setPercolatorFeatures(
      $objMascotOptions,
      '',
      $objResFile->params->getML_ADAPTER_PARAM
    );
    $vecFiles = $objResFile->getPercolatorFileNames;
    unless ($vecFiles 
      && $$vecFiles[$msparser::ms_mascotresfilebase::PERCOLATOR_OUTPUT_TARGET]
      && -e $$vecFiles[$msparser::ms_mascotresfilebase::PERCOLATOR_OUTPUT_TARGET]
      && $$vecFiles[$msparser::ms_mascotresfilebase::PERCOLATOR_OUTPUT_DECOY]
      && -e $$vecFiles[$msparser::ms_mascotresfilebase::PERCOLATOR_OUTPUT_DECOY]) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=POP files don't exist\n";
      exit 1;
    }
    
  # Read each POP file and write to STDOUT in MIME format
    unless (open(TARGETFILE, $$vecFiles[$msparser::ms_mascotresfilebase::PERCOLATOR_OUTPUT_TARGET])) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=unable to open target.pop file for reading\n";
      exit 1;
    }
    unless (open(DECOYFILE, $$vecFiles[$msparser::ms_mascotresfilebase::PERCOLATOR_OUTPUT_DECOY])) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=unable to open decoy.pop file for reading\n";
      exit 1;
    }
    print "Content-Type: multipart/form-data; boundary=\"-------------------------7da27721420678\"\n\n";
    print "---------------------------7da27721420678\n";
    print "Content-Disposition: form-data; name=\"target.pop\"; filename=\""
      . $$vecFiles[$msparser::ms_mascotresfilebase::PERCOLATOR_OUTPUT_TARGET]
      . "\"\n\n";
    while (<TARGETFILE>) {
      print $_;
    }
    print "---------------------------7da27721420678\n";
    print "Content-Disposition: form-data; name=\"decoy.pop\"; filename=\""
      . $$vecFiles[$msparser::ms_mascotresfilebase::PERCOLATOR_OUTPUT_DECOY]
      . "\"\n\n";
    while (<DECOYFILE>) {
      print $_;
    }
    print "---------------------------7da27721420678--\n";
    close(TARGETFILE);
    close(DECOYFILE);

  } elsif (defined($thisScript->param('version'))) {

    # Temporary workaround to bypass version check in Mascot node in Proteome Discoverer
    print "../bin/ms-searchcontrol.exe - Version 2.9.0\n";
    exit 0;

  # Get version information for ms-searchcontrol.exe
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --version |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    print @temp;

  } elsif (defined($thisScript->param('version_all'))) {

  # Get version information for key binary files
    my @exe_list = qw(../bin/ms-compress.exe
        ../bin/ms-monitor.exe
        ../bin/ms-makesearchlog.exe
        ../bin/ms-searchcontrol.exe
        ../cgi/ms-geterror.exe
        ../cgi/nph-mascot.exe
        ../cgi/nph-mascot?.exe
        ../x-cgi/ms-getseq.exe
        ../x-cgi/ms-gettaxonomy.exe
        ../x-cgi/ms-status.exe
        ../x-cgi/ms-review.exe
        ../x-cgi/ms-config.exe);
    foreach my $exe (@exe_list) {
      while (my $fileName = glob($exe)) {
        if (open(SOCK, "\"$fileName\" --version |")) {
          my @temp = <SOCK>;
          close(SOCK);
          foreach (@temp) {
            if (/version/i) {
              print $_ . "\n";
              last;
            }
          }
        }
      }
    }

  } elsif (defined($thisScript->param('nice_job'))) {
    
  # security is handled by ms-searchcontrol.exe

    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

    my $nice_value = 0;
    if (defined($thisScript->param('nice_value'))) {
      $nice_value = 0 + $thisScript->param('nice_value');
    }

  # Call ms-searchcontrol.exe to set search priority
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --nice_job --taskID ".common_subs::shellQuote($taskID)." --nice ".common_subs::shellQuote($nice_value)." $sessionID_parameter |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    print @temp;

  } elsif (defined($thisScript->param('querydata'))) {
    
    my(@queryNums, $query_number, $fileIn);
    
    if ($thisScript->param('query_number')) {
      $query_number = $thisScript->param('query_number');
      # The accepted format covers comma-separated values and ranges:
      #  query_number=a[[-b][,c[-d]]...]
      @queryNums = map { 
        my ($a, $b) = m/^(\d+)(?:-(\d+))?$/;
        if (defined $a) {
          if (defined $b) {
            $a .. $b
          } else {
            $a
          }
        } else {
          ()
        }
      } split /,/, $query_number;
    } else {
      print "problem=no query_number value(s) supplied for querydata\n";
      exit 1;
    }
    
    if (defined($thisScript->param('file'))) {
      $fileIn = $thisScript->param('file');
      $fileIn = undef unless $fileIn =~ m/^[^\x00-\x1f]+$/;
    } else {
      if ($taskID < 1) {
        print "problem=Invalid task ID '$taskID'\n";
        exit 1;
      }
      unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --result_file_name --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
        print "problem=Could not open socket to ms-searchcontrol.exe\n";
        exit 1;
      }
      my @temp = <SOCK>;
      close(SOCK);
      $fileIn = "";
      foreach (@temp) {
        if (/^filename=(.+)/i) {
          $fileIn = $1;
          last;
        }
      }
    }
    
  # May need to decompress the result file
    if ($fileIn) {
      unless (-e common_subs::decompress($fileIn)) {
        print "problem=result file does not exist\n";
        exit 1;
      }
    } else {
      print "problem=no result file name\n";
      exit 1;
    }
    
  # create an configuration options object from mascot.dat
    my $objDatFile = new msparser::ms_datfile("../config/mascot.dat");
  # abort on fatal error. Note that isValid() is true if there are only warnings
    unless ($objDatFile->isValid) {
      my $errorList = "problem=Mascot Parser error(s) at 5:\n";
      my $objErr = $objDatFile->getErrorHandler();
      for (my $i = 1; $i <= $objErr->getNumberOfErrors(); $i++) {
        $errorList .= $objErr->getErrorNumber($i) 
          . ": " . common_subs::htmlEscape($objErr->getErrorString($i)) . "\n";
      }
      print $errorList;
      exit 1;
    }
    my $objMascotOptions = $objDatFile->getMascotOptions;

  # see whether to use ms_mascotresfilebase::RESFILE_USE_CACHE flag
    my $resfileFlags;
    if (common_subs::useResfileCache($objMascotOptions, 'client.pl')) {
      $resfileFlags = $msparser::ms_mascotresfilebase::RESFILE_USE_CACHE;
    } else {
      $resfileFlags = $msparser::ms_mascotresfilebase::RESFILE_NOFLAG;
    }
    my $resfileCacheDirectory = $objMascotOptions->getCacheDirectory;

  # create Mascot Parser result file object
    my $objResFile = _open_resfile(common_subs::decompress($fileIn), 
                                                $resfileFlags, 
                                                $resfileCacheDirectory);
  
  # abort on fatal error. Note that isValid() remains true if there are only warnings
    unless ($objResFile->isValid) {
      print "problem=error on trying to create result file object\n";
      exit 1;
    }

  # Mascot security
    my $objParams = new msparser::ms_searchparams($objResFile);
    if ($session->isValid) {
      unless ($session->canResultsFileBeViewed($objParams->getUSERID)) {
        print "problem=Not permitted by Mascot security\n";
        exit 1;
      }
    } else {
      print "problem=Not logged in to Mascot security\n";
      exit 1;
    }
  
  # output the selected queries
    outputQueries(\$objResFile, \@queryNums);

  } elsif (defined($thisScript->param('user_params'))) {
    
    if ($taskID < 1) {
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

  # May need to decompress the result file
    unless (open(SOCK, "\"../bin/ms-searchcontrol.exe\" --result_file_name --taskID ".common_subs::shellQuote($taskID)." $sessionID_parameter |")) {
      print "problem=Could not open socket to ms-searchcontrol.exe\n";
      exit 1;
    }
    my @temp = <SOCK>;
    close(SOCK);
    my $fileIn = "";
    foreach (@temp) {
      if (/^filename=(.+)/i) {
        $fileIn = $1;
        last;
      }
    }
    if ($fileIn) {
      if (-e common_subs::decompress($fileIn)) {
      # Remove the .flg file so that results don't get emailed.
        unlink common_subs::decompress($fileIn) . ".flg";
      # Also remove the batch log file if it exists.
        unlink common_subs::decompress($fileIn) . ".log";
      } else {
        print "problem=result file does not exist\n";
        exit 1;
      }
    } else {
      print "problem=no result file name\n";
      exit 1;
    }

  # create an configuration options object from mascot.dat
    my $objDatFile = new msparser::ms_datfile("../config/mascot.dat");
  # abort on fatal error. Note that isValid() is true if there are only warnings
    unless ($objDatFile->isValid) {
      my $errorList = "problem=Mascot Parser error(s) at 6:\n";
      my $objErr = $objDatFile->getErrorHandler();
      for (my $i = 1; $i <= $objErr->getNumberOfErrors(); $i++) {
        $errorList .= $objErr->getErrorNumber($i) 
          . ": " . common_subs::htmlEscape($objErr->getErrorString($i)) . "\n";
      }
      print $errorList;
      exit 1;
    }
    my $objMascotOptions = $objDatFile->getMascotOptions;

  # see whether to use ms_mascotresfilebase::RESFILE_USE_CACHE flag
    my $resfileFlags;
    if (common_subs::useResfileCache($objMascotOptions, 'client.pl')) {
      $resfileFlags = $msparser::ms_mascotresfilebase::RESFILE_USE_CACHE;
    } else {
      $resfileFlags = $msparser::ms_mascotresfilebase::RESFILE_NOFLAG;
    }
    my $resfileCacheDirectory = $objMascotOptions->getCacheDirectory;

  # create Mascot Parser result file object
    my $objResFile = _open_resfile(common_subs::decompress($fileIn), 
                                                $resfileFlags, 
                                                $resfileCacheDirectory);
  
  # abort on fatal error. Note that isValid() remains true if there are only warnings
    unless ($objResFile->isValid) {
      print "problem=error on trying to create result file object\n";
      exit 1;
    }
  
  # Mascot security
    my $objParams = new msparser::ms_searchparams($objResFile);
    if ($session->isValid) {
      unless ($session->canResultsFileBeViewed($objParams->getUSERID)) {
        print "problem=Not permitted by Mascot security\n";
        exit 1;
      }
    } else {
      print "problem=Not logged in to Mascot security\n";
      exit 1;
    }

  # Print out any user parameters with defined values, that is USER00 to USER12 and anything
  # beginning with underscore unless they match the following reserved patterns
  # _integra_*
  # _daemon_*
  # _distiller_*
  # _server_*

    my $keys = msparser::VectorString->new;
    my $vals = msparser::VectorString->new;
    $objResFile->getSearchParametersKeyValues($keys, $vals);
    for my $i (0 .. $keys->size-1) {
      my $nextKey = $keys->get($i);
      if ($nextKey =~ /^_/i || $nextKey =~ /^user\d\d$/i) {
        if ($nextKey !~ /^_integra_/i && $nextKey !~ /^_daemon_/i && $nextKey !~ /^_distiller_/i && $nextKey !~ /^_server_/i) {
          my $nextValue = $vals->get($i);
          $nextKey =~ s/\s*$//;
          if ($nextValue) {
              print $nextKey . "=" . $nextValue . "\n";
          }
        }
      }
    }

  } elsif (defined($thisScript->param('quant_xml'))) {

    if ($session->isValid) {
      unless ($session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIGUSINGMSSTATUS)
        || $session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIG)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Not permitted by Mascot security\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Not logged in to Mascot security\n";
      exit 1;
    }

  # To obtain quantitation.xml file from ms-status.exe
    if (system("../x-cgi/ms-status.exe", "Show=MS_QUANTITATIONXML", "sessionID=$sessionID")) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Error return from ms-status.exe\n";
      exit 1;
    }

  } elsif (defined($thisScript->param('quant_method'))) {

    if ($session->isValid) {
      unless ($session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIGUSINGMSSTATUS)
        || $session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIG)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Not permitted by Mascot security\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Not logged in to Mascot security\n";
      exit 1;
    }

  # To obtain pretty-printed (HTML) quantitation method summary from ms-config.exe
    if (defined($thisScript->param('file'))) {
      my $fileIn = $thisScript->param('file');
      $fileIn = '' unless $fileIn =~ m/^[^\x00-\x1f]+$/;

      unless (-e common_subs::decompress($fileIn)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Result file doesn't exist\n";
        exit 1;
      }
      if (system("../x-cgi/ms-config.exe", "printmethod", "resfile=" . common_subs::decompress($fileIn))) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Error return from ms-config.exe\n";
        exit 1;
      }
    } elsif (defined($thisScript->param('method'))) {
      my $method = $thisScript->param('method');
      if (system("../x-cgi/ms-config.exe", "printmethod=$method")) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Error return from ms-config.exe\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Second argument required, either file or method\n";
      exit 1;
    }

  } elsif (defined($thisScript->param('modification'))) {

    if ($session->isValid) {
      unless ($session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIGUSINGMSSTATUS)
        || $session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIG)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Not permitted by Mascot security\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Not logged in to Mascot security\n";
      exit 1;
    }

    if (defined($thisScript->param('mod_name')) && defined($thisScript->param('file'))) {
      my $print_mod_name = 'printmod='.$thisScript->param('mod_name');
      my $fileIn = $thisScript->param('file');
      $fileIn = '' unless $fileIn =~ m/^[^\x00-\x1f]+$/;

      unless (-e common_subs::decompress($fileIn)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Result file doesn't exist\n";
        exit 1;
      }
      if (system("../x-cgi/ms-config.exe", $print_mod_name, "resfile=" . common_subs::decompress($fileIn))) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Error return from ms-config.exe\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Modification name and file path required as arguments\n";
      exit 1;
    }

  } elsif (defined($thisScript->param('crosslinking_xml'))) {

    if ($session->isValid) {
      unless ($session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIGUSINGMSSTATUS)
        || $session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIG)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Not permitted by Mascot security\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Not logged in to Mascot security\n";
      exit 1;
    }

  # To obtain crosslinking.xml file from ms-status.exe
    if (system("../x-cgi/ms-status.exe", "Show=MS_CROSSLINKINGXML", "sessionID=$sessionID")) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Error return from ms-status.exe\n";
      exit 1;
    }

  } elsif (defined($thisScript->param('crosslinking_method'))) {

    if ($session->isValid) {
      unless ($session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIGUSINGMSSTATUS)
        || $session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIG)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Not permitted by Mascot security\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Not logged in to Mascot security\n";
      exit 1;
    }

  # To obtain pretty-printed (HTML) quantitation method summary from ms-config.exe
    if (defined($thisScript->param('file'))) {
      my $fileIn = $thisScript->param('file');
      $fileIn = '' unless $fileIn =~ m/^[^\x00-\x1f]+$/;

      unless (-e common_subs::decompress($fileIn)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Result file doesn't exist\n";
        exit 1;
      }
      if (system("../x-cgi/ms-config.exe", "print_xlink_method", "resfile=" . common_subs::decompress($fileIn))) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Error return from ms-config.exe\n";
        exit 1;
      }
    } elsif (defined($thisScript->param('method'))) {
      my $method = $thisScript->param('method');
      if (system("../x-cgi/ms-config.exe", "print_xlink_method=$method")) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Error return from ms-config.exe\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Second argument required, either file or method\n";
      exit 1;
    }

  } elsif (defined($thisScript->param('linker'))) {

    if ($session->isValid) {
      unless ($session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIGUSINGMSSTATUS)
        || $session->isPermitted($msparser::ms_security_tasks::SECTASK_VIEWCONFIG)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Not permitted by Mascot security\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Not logged in to Mascot security\n";
      exit 1;
    }

    if (defined($thisScript->param('mod_name')) && defined($thisScript->param('file'))) {
      my $print_mod_name = 'printlinker='.$thisScript->param('mod_name');
      my $fileIn = $thisScript->param('file');
      $fileIn = '' unless $fileIn =~ m/^[^\x00-\x1f]+$/;

      unless (-e common_subs::decompress($fileIn)) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Result file doesn't exist\n";
        exit 1;
      }
      if (system("../x-cgi/ms-config.exe", $print_mod_name, "resfile=" . common_subs::decompress($fileIn))) {
        print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
        print "problem=Error return from ms-config.exe\n";
        exit 1;
      }
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Modification name and file path required as arguments\n";
      exit 1;
    }

  } elsif (defined($thisScript->param('osname'))) {
    
    print $^O;

  } elsif (defined($thisScript->param('db_active'))) {

    my $objDatFile = new msparser::ms_datfile("../config/mascot.dat");
    if (!$objDatFile->isValid) {
      print "problem=" . $objDatFile->getLastErrorString() . "\n";
      exit 1;
    }
  # retrieve Databases section content
    my $objDatabases = $objDatFile->getDatabases;
  # check if there is actually a 'Databases' section in the file 
    if ($objDatabases->isSectionAvailable) {
      for(my $i = 0; $i < $objDatabases->getNumberOfDatabases; $i++) {
        if (lc($objDatabases->getDatabase($i)->getName) eq lc($thisScript->param('db_name'))) {
          if ($objDatabases->getDatabase($i)->isActive()) {
            print "true";
            last;
          }
        }
      }
    } else {
      print "problem=No databases secion in mascot.dat\n";
      exit 1;
    }

  } elsif (defined($thisScript->param('fileset_create'))) {

    my $session = new msparser::ms_session(FILESET_TIMEOUT, 'fileset');

    if (not $session->isValid) {
      print "problem=Cannot create a session: ", $session->getLastErrorString, " \n";
      exit 1;
    }

    print $session->getID;

  } elsif (defined($thisScript->param('fileset_append'))) {

    my $fileset = $thisScript->param('fileset_id');

    if (not $fileset) {
      print "problem=Parameter 'fileset_id' is empty\n";
      exit 1;
    }

    my $session = new msparser::ms_session($fileset);

    if (not $session->isValid) {
      print "problem=Cannot open session: ", $session->getLastErrorString, "\n";
      exit 1;
    }

    my @files = ();
  # Bug 12292: Perl 5.20 and later require using CGI::multi_param() in list
  # context, but this method doesn't exist in Perl 5.16 or earlier. So
  # we need to detect it first.
    if ($thisScript->can('multi_param')) {
      @files = $thisScript->multi_param('file');
    } else {
      @files = $thisScript->param('file');
    }
    my $datfile = msparser::ms_datfile->new();
    my $datadir = $datfile->getMascotOptions->getInterFileRelPath();

    for my $file (@files) {
      my $filepath = $file;
      next unless $file =~ m/^[^\x00-\x1f]+$/;

      if ($file !~ m#/#) {
          $filepath = join('/', $datadir, $file);
      }

      $filepath = common_subs::decompress($filepath);

      if (not -e $filepath) {
        print "problem=File '".$file."' does not exist\n";
        exit 1;
      }
    }

    my $set = $session->getStringParam('Fileset');
    $session->saveStringParam('Fileset',
      join("\n", length($set) == 0 ? () : $set, @files)
    );

    print scalar(@files), " added\n";

  } elsif (defined($thisScript->param('fileset_remove'))) {

    my $fileset = $thisScript->param('fileset_id');

    if (not $fileset) {
      print "problem=Parameter 'fileset_id' is empty\n";
      exit 1;
    }

    my $session = new msparser::ms_session($fileset);

    if (not $session->isValid) {
      print "problem=Cannot open session: ", $session->getLastErrorString, "\n";
      exit 1;
    }

    my @files = ();
  # Bug 12292: Perl 5.20 and later require using CGI::multi_param() in list
  # context, but this method doesn't exist in Perl 5.16 or earlier. So
  # we need to detect it first.
    if ($thisScript->can('multi_param')) {
      @files = $thisScript->multi_param('file');
    } else {
      @files = $thisScript->param('file');
    }
    my %files = map { $_ => 1 } @files;
    my @originals = split /\n/, $session->getStringParam('Fileset');
    my @new = grep { $files{$_} ? () : $_ } @originals;

    $session->saveStringParam('Fileset', join("\n", @new));

    my $num = scalar(@originals) - scalar(@new);

    print "$num removed\n";

  } elsif (defined($thisScript->param('fileset_list'))) {

    my $fileset = $thisScript->param('fileset_id');

    if (not $fileset) {
      print "problem=Parameter 'fileset_id' is empty\n";
      exit 1;
    }

    my $session = new msparser::ms_session($fileset);

    if (not $session->isValid) {
      print "problem=Cannot open session: ", $session->getLastErrorString, "\n";
      exit 1;
    }

    my $set = $session->getStringParam('Fileset');
    print $set, "\n";

  } elsif (defined($thisScript->param('download_distiller_xml'))) {

    my ($taskID, $sessionID, $session, $sessionID_parameter) = _retrieve_common_parameters($thisScript);

    if (not $session->isPermitted($msparser::ms_security_tasks::SECTASK_DOWNLOAD_DISTILLER_XML)) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Not permitted by Mascot Security\n";
      exit 1;
    }

    if ($taskID < 1) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Invalid task ID '$taskID'\n";
      exit 1;
    }

    my $resfilepath = '';

  # Call ms-searchcontrol.exe to get result file name.
  # ms-searchcontrol.exe will validate the session as well.
    if (open(my $pipe, '"../bin/ms-searchcontrol.exe" --result_file_name --taskID '.common_subs::shellQuote($taskID).' '.$sessionID_parameter.' |')) {
      while (<$pipe>) {
        if (/^filename=(.+)/i) {
          chomp($resfilepath = $1);
          last;
        }
      }
      1 while (<$pipe>); # flush the pipe so that zombies won't arise
    } else {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Could not open socket to ms-searchcontrol.exe: $! ($^E)\n";
      exit 1;
    }

    if ($resfilepath eq '') {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=no result file name\n";
      exit 1;
    }

    if (not -e $resfilepath) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=result file does not exist\n";
      exit 1;
    }

    my $file_md5 = do {
        my $str = $thisScript->param('distiller_xml_filename');

        if (not defined $str or $str eq '') {
          print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
          print "problem=distiller_xml_filename must be given\n";
          exit 1;
        }

        _distiller_xml_filename_to_md5($str)
    };

    if (not defined $file_md5) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Invalid distiller_xml_filename\n";
      exit 1;
    }

    my $xmlfilepath = _resfilepath_to_distiller_xml_filepath($resfilepath, $file_md5);

    if (not -e $xmlfilepath) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Distiller XML file does not exist\n";
      exit 1;
    }

    my ($fh, $errormsg, undef, $warn_errormsg) = _acquire_read_lock($xmlfilepath);

    if (not $fh) {
      print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
      print "problem=Distiller XML cannot be read: $errormsg\n";
      warn $warn_errormsg;
      exit 1;
    }

    binmode(STDOUT);

    my ($basename) = $xmlfilepath =~ m#([^/]+)$#;

    print $thisScript->header( 
        -type => "text/xml", -charset => "utf-8", -attachment => $basename,
    );

    while (1) {
      my $len = read($fh, my $buffer, 1048576);
      if (not defined $len) {
        warn "Problem while reading Distiller XML file '$xmlfilepath': $! ($^E)";
        last;
      }
      if ($len == 0) {
        last;
      }
      print $buffer;
    }

    close($fh);

  } elsif (defined($thisScript->param('recent_distiller_xml'))) {

    my ($taskID, $sessionID, $session, $sessionID_parameter) = _retrieve_common_parameters($thisScript);

    if (not $session->isPermitted($msparser::ms_security_tasks::SECTASK_DOWNLOAD_DISTILLER_XML)) {
      print "problem=Not permitted by Mascot Security\n";
      exit 1;
    }

    my $timestamp_from = $thisScript->param('from');
    $timestamp_from = 0 if (not($timestamp_from) or $timestamp_from < 0);

    my ($fh, $errormsg, $file_exists, $warn_errormsg) = _acquire_read_lock(DISTILLER_XML_LOGFILE);

    if (not $fh) {
      if (not $file_exists) {
        print "success=No new messages since timestamp $timestamp_from\n";
        exit 0;
      }
      print "problem=Distiller XML log file cannot be read: $errormsg\n";
      warn $warn_errormsg;
      exit 1;
    }

    my $stat = stat($fh);

    if (not defined $stat) {
      print "problem=Distiller XML log file cannot be read: $! ($^E)\n";
      exit 1;
    }

    if ($stat->mtime <= $timestamp_from) {
      print "success=No new messages since timestamp $timestamp_from\n";
      exit 0;
    }

    binmode(STDOUT);

    if ($timestamp_from) {
      print "success=New messages since timestamp $timestamp_from:\n";
    } else {
      print "success=All log messages:\n";
    }
    print '# ', join("\t", @{ DISTILLER_MSGHEADER() }), "\n";

    while (<$fh>) {
      chomp;
      my ($timestamp) = m#^(\d+)#;
      next if (not defined $timestamp or $timestamp eq '');
      next if $timestamp <= $timestamp_from;
      print $_, "\n";
    }

    close($fh);

  } elsif (defined($thisScript->param('distiller_xml_exists'))) {

    # sessionID is ignored. task_id can be an array.

    my %task_id_query = ();

    for ($thisScript->param('task_id')) {
      my ($id) = m#^(\d+)$#;
      if ($id) {
        $task_id_query{$id} = 0;
      }
    }

    if (not keys %task_id_query) {
      print "problem=No valid task_id given\n";
      exit 1;
    }

    my ($fh, $errormsg, $file_exists, $warn_errormsg) = _acquire_read_lock(DISTILLER_XML_LOGFILE);

    if (not $fh) {
      if (not $file_exists) {
        for my $task_id (keys %task_id_query) {
          printf "%d=n\n", $task_id;
        }
        exit 0;
      }
      print "problem=Distiller XML log file cannot be read: $errormsg\n";
      warn $warn_errormsg;
      exit 1;
    }

    my %parsed_ids = %task_id_query;

    while (<$fh>) {
      last if not keys %task_id_query;
      chomp;
      my @fields = split /\t/;
      my ($task_id) = $fields[3] =~ m#^(\d+)#;
      next if not $task_id;
      my ($success) = $fields[2] =~ m#^([YN])#;
      next if (not defined($success) or $success eq '');
      if (exists $task_id_query{$task_id}) {
        if ($success eq 'Y') {
          $parsed_ids{$task_id} = 1;
          delete $task_id_query{$task_id};
        }
      }
    }

    for my $task_id (sort { $a <=> $b } keys %parsed_ids) {
      printf "%d=%1s\n", $task_id, $parsed_ids{$task_id} ? 'y' : 'n';
    }

  } else {
    print "problem=Invalid keyword argument supplied to client.pl\n";
    exit 1;

  }
  
  exit 0;
} else {
  print $thisScript->header( -type => "text/plain", -charset => "utf-8" );
  print "problem=Unknown request method ", $thisScript->request_method, "\n";
  exit 0;
}

sub _retrieve_common_parameters {
  my ($thisScript) = @_;

  my $taskID = '0';
  if (defined($thisScript->param('task_id'))) {
    ($taskID) = $thisScript->param('task_id') =~ m#^(\d+)$#;
    $taskID = '0' if not $taskID;
  }

  my $sessionID = '';
  my $sessionID_parameter = '';
  if ($thisScript->param('sessionID')) {
    $sessionID = $thisScript->param('sessionID');
    # Bug 14532: Assign the URL parameter here, so that it's passed to
    # ms-searchcontrol.exe commands. We need to do this even if the
    # session is invalid.
    $sessionID_parameter = "--sessionID ".common_subs::shellQuote($sessionID);
  }

  my $session = new msparser::ms_session($sessionID);
  if ($session->isValid) {
    $sessionID = $session->getID();
    # Bug 14532: Reassign the parameter here now that we know the
    # session is valid, so that ms-searchcontrol.exe uses the
    # identical session.
    $sessionID_parameter = "--sessionID ".common_subs::shellQuote($sessionID);
  }

  return ($taskID, $sessionID, $session, $sessionID_parameter);
}

###############################################################################
# &outputQueries()
# output XML formatted queries
# $_[0] reference to result file object
# $_[1] reference to list of query numbers
###############################################################################

sub outputQueries {

    my $export_format = "XML";
    
    print "<queries>\n";
    
    my $string = "";
    for my $queryNum (@{ $_[1] }) {
        my $objQuery = new msparser::ms_inputquery(${ $_[0] }, $queryNum);
        if ($export_format eq "CSV") {
            $string .= $queryNum . ",";
        } elsif ($export_format eq "XML") {
            $string .= "<query number=\"$queryNum\">,";
        }
        $string .= formatElement($export_format, "", 1, "StringTitle", 
            &noXmlTag($objQuery->getStringTitle(1))) . ",";
        my @qualifiers;
        if ($objQuery->getPepTolString()) {
            if ($export_format eq "XML") {
                push @qualifiers, formatElement($export_format, "", 1, "qual_tol", $objQuery->getPepTolString());
            }	else {
                push @qualifiers, $objQuery->getPepTolString();
            }
        }
        for (my $j = 1; $j <= 20; $j++) {
            if ($objQuery->getSeq($j)) {
                if ($export_format eq "XML") {
                    push @qualifiers, formatElement($export_format, "", 1, "qual_seq", "seq(" . $objQuery->getSeq($j) . ")", "number", $j);
                }	else {
                    push @qualifiers, "seq(" . $objQuery->getSeq($j) . ")";
                }
            } else {
                last;
            }
        }
        for (my $j = 1; $j <= 20; $j++) {
            if ($objQuery->getComp($j)) {
                if ($export_format eq "XML") {
                    push @qualifiers, formatElement($export_format, "", 1, "qual_comp", "comp(" . $objQuery->getComp($j) . ")", "number", $j);
                }	else {
                    push @qualifiers, "comp(" . $objQuery->getComp($j) . ")";
                }
            } else {
                last;
            }
        }
        for (my $j = 1; $j <= 20; $j++) {
            if (my $local_tag_string = $objQuery->getTag($j)) {
                $local_tag_string =~ /([et]),(.*)/i;
                if ($1 eq "e") {
                    $local_tag_string = "etag($2)";
                } else {
                    $local_tag_string = "tag($2)";
                }
                if ($export_format eq "XML") {
                    push @qualifiers, formatElement($export_format, "", 1, "qual_tag", $local_tag_string, "number", $j);
                }	else {
                    push @qualifiers, $local_tag_string;
                }
            } else {
                last;
            }
        }
        if ($export_format eq "XML") {
            if (@qualifiers) {
                $string .= join(",", @qualifiers) . ",";
            }
        } else {
            $string .= formatElement($export_format, "", 1, "pep_qual", join(" ", @qualifiers)) . ",";
        }
        $string .= formatElement($export_format, "", 0, "TotalIonsIntensity", $objQuery->getTotalIonsIntensity()) . ",";
        $string .= formatElement($export_format, "", 0, "NumVals", $objQuery->getNumVals()) . ",";
        $string .= formatElement($export_format, "", 1, "StringIons1", $objQuery->getStringIons1()) . ",";
        $string .= formatElement($export_format, "", 1, "StringIons2", $objQuery->getStringIons2()) . ",";
        $string .= formatElement($export_format, "", 1, "StringIons3", $objQuery->getStringIons3());
        if ($export_format eq "XML") {
            $string .= ",</query>";
        }
        $string .= "\n";
    }
    
    chop $string;
    if ($export_format eq "XML") {
  	$string =~ s/>,</>\n</g;
    }
    if ($string) {
        print $string . "\n";
    }

    if ($export_format eq "XML") {
  	print "</queries>\n";
    }	

}

###############################################################################
# &formatElement()
# returns formatted element
#
# $_[0] "XML" or "CSV" or ...
# $_[1] CSV label, empty if none required
# $_[2] 1 if text element requires double quotes for CSV
# $_[3] xml element name
# $_[4] element value
# $_[5] attribute name for XML, empty if none
# $_[6] attribute value for XML, empty if none
###############################################################################

sub formatElement {
    
    my($format, $label, $need_quotes, $name, $value, $attrib_name, $attrib_value) = @_;
    
    my $return = "";
    
    if ($format eq "XML") {
        $return .= "<$name";
        if ($attrib_name) {
            $return .= " $attrib_name=\"$attrib_value\"";
        }
        if ($value eq "") {
            $return .= "/>";
        } else {
            $return .= ">$value</$name>";
  	}
    } elsif ($format eq "CSV") {
        if ($label) {
            $return .= "$label,";
        }
        if ($need_quotes) {
            $return .= "\"";
        }
        $return .= $value;
        if ($need_quotes) {
            $return .= "\"";
        }
    }
    
    return $return;

}

###############################################################################
# &noXmlTag()
# returns de-tagged string
# $_[0] string which may contain XML illegal characters
###############################################################################

sub noXmlTag {

  my $temp = shift;

  $temp =~ s/&/&amp;/g;
  $temp =~ s/</&lt;/g;
  $temp =~ s/>/&gt;/g;

  return $temp;

}

sub _open_resfile {
    my ($decompressed_fileIn, $resfileFlags, $resfileCacheDirectory) = @_;

    return  msparser::ms_mascotresfilebase::createResfile(
      $decompressed_fileIn, 0, "", $resfileFlags, $resfileCacheDirectory
    );
}

sub _log_distiller_upload_success {
    my ($taskID, $filepath, $computer_name, $original_directory) = @_;

    my ($fh) = _acquire_append_lock(DISTILLER_XML_LOGFILE);

    if (not $fh) {
        warn "Upload succeeded but could not log success";
        return;
    }

    _possibly_output_header($fh);

    printf $fh DISTILLER_MSGFMT."\n",
        time,
        scalar(localtime(time)),
        'Y',
        $taskID,
        $filepath,
        _sanitise($computer_name),
        _sanitise($original_directory),
        'Success',
        ;

    return;
}

sub _possibly_output_header {
    my ($fh) = @_;

    if (-s $fh == 0) {
        # Brand new file!
        print $fh '# ', join("\t", @{ DISTILLER_MSGHEADER() }), "\n";
    }
}

sub _sanitise {
    my ($str) = @_;
    return '(none)' if (not defined $str or $str eq '');
    $str =~ s/[\t\n\r\0]/ /g;
    return $str;
}

sub _log_distiller_upload_failure {
    my ($taskID, $filepath, $computer_name, $original_directory, $msg) = @_;

    my ($fh) = _acquire_append_lock(DISTILLER_XML_LOGFILE);

    if (not $fh) {
        warn "Error to log is: $msg";
        return;
    }

    _possibly_output_header($fh);

    printf $fh DISTILLER_MSGFMT."\n",
        time,
        scalar(localtime(time)),
        'N',
        $taskID,
        $filepath,
        _sanitise($computer_name),
        _sanitise($original_directory),
        _sanitise($msg),
        ;

    return;
}

sub _acquire_read_lock {
    my ($filepath) = @_;

    if (not -e $filepath) {
        return (undef, undef, 0);
    }

    open(my $fh, '<', $filepath) or do {
        return (undef, "$! ($^E)", 1, "Cannot open '$filepath' for reading: $! ($^E)");
    };

    flock($fh, LOCK_SH) or do {
        return (undef, "$! ($^E)", 1, "Cannot lock '$filepath': $! ($^E)");
    };

    seek($fh, 0, 0) or do {
        return (undef, "$! ($^E)", 1, "Cannot seek '$filepath': $! ($^E)");
    };
    
    return $fh;
}

sub _acquire_append_lock {
    my ($filepath) = @_;

    open(my $fh, '>>', $filepath) or do {
        warn "Cannot open '$filepath' for appending: $! ($^E)";
        return (undef, "$! ($^E)");
    };

    flock($fh, LOCK_EX) or do {
        warn "Cannot lock '$filepath': $! ($^E)";
        return (undef, "$! ($^E)");
    };

    seek($fh, 0, 2) or do {
        warn "Cannot seek '$filepath': $! ($^E)";
        return (undef, "$! ($^E)");
    };

    return $fh;
}

sub _acquire_write_lock {
    my ($filepath) = @_;

    my $mode;

    if (-e $filepath) {
        $mode = '+<';
    } else {
        $mode = '+>';
    }

    open(my $fh, $mode, $filepath) or do {
        return (undef, "$! ($^E)", -e $filepath, "Cannot open '$filepath' for writing: $! ($^E)");
    };

    flock($fh, LOCK_EX) or do {
        return (undef, "$! ($^E)", 1, "Cannot lock '$filepath': $! ($^E)");
    };

    return $fh;
}

sub _resfilepath_to_distiller_xml_filepath {
    my ($resfilepath, $file_md5) = @_;
    $resfilepath =~ s/(\.[^.]+)$/.$file_md5.quant.xml/;
    return $resfilepath;
}

sub _distiller_xml_filename_to_md5 {
    my ($str) = @_;
    return if (not defined $str or $str eq '');
    my ($md5) = $str =~ m#\.([a-zA-Z0-9]+)\.quant\.xml$#;
    return if (not defined $md5 or $md5 eq '');
    return $md5;
}


###############################################################################
# my @lines = read_10_lines($pipe_in);
#
# Read up to 10 lines from the pipe. The lines are read in normal line mode
# and the string elements in @lines include native line endings. If nothing
# was read (e.g. invalid pipe or command printed nothing to stdout), then
# @lines is empty.
#
# $_[0] pipe to an external command opened with open()
###############################################################################
sub read_10_lines {
    my ($pipe_in) = @_;

    my @lines;
    
    for (1 .. 10) {
        my $line = <$pipe_in>;
        
        if (not defined $line) {
            # eof
            last;
        }

        # Note, line ending character(s) are kept.
        push @lines, $line;
    }

    return @lines;
}

###############################################################################
# my $errnum = parse_ms_searchcontrol_error(\@lines);
#
# Read each line of text (ignoring line endings). If it starts with
# "searchcontrol-error=", parse and return the integer error number. If nothing
# was parsed, return undef.
#
# $_[0] array reference of lines of text
###############################################################################
sub parse_ms_searchcontrol_error {
    my ($lines) = @_;

    for (@$lines) {
        if (/^searchcontrol-error=(\d+)/) {
            return $1;
        }
    }

    return undef;
}

###############################################################################
# my @lines = map_ms_searchcontrol_error_to_description($errnum);
#
# Given an ms-searchcontrol.exe error number, return one or more lines of
# text that are suitable as the output of client.pl. The function always
# returns something. If $errnum isn't a known error number, it returns e.g.:
#
#  searchcontrol-error=13
#  problem=Error return from ms-searchcontrol.exe
###############################################################################
sub map_ms_searchcontrol_error_to_description {
    my ($errnum) = @_;

    if (defined $errnum) {
        # Error numbers are defined in
        # matrixlib/ms_uniquetask.hpp. Only a few should be
        # possible here.

        if ($errnum == 8) {
            return (
                'problem=Results file cannot be found or is empty',
            );
        }
        if ($errnum == 10) {
            return (
                'problem=Not logged in to Mascot security',
            );
        }
        if ($errnum == 11) {
            return (
                'problem=Not permitted by Mascot security',
            );
        }
    }

    return (
        sprintf('searchcontrol-error=%d', $errnum),
        'problem=Error return from ms-searchcontrol.exe',
    );
}


###############################################################################
# my $status = refiner_status($resfilepath);
#
# Call refine_results_with_ml.pl to find the status of refining.
# Return value is one of:
#
# not_created
# created
# creating=%d%%
# error=...
#
# Note, these all end with a newline (CRLF or LF).
###############################################################################
sub refiner_status {
    my ($resfilepath) = @_;

    my $perl_path = perl_path();

    my $pipe_in;
    unless (open($pipe_in, "\"$perl_path\" ../bin/refine_results_with_ml.pl --status --resfile ".common_subs::shellQuote($resfilepath).' |')) {
        return "error=Could not open socket to refine_results_with_ml.pl\n";
    }

    my @output = <$pipe_in>;

    if (@output) {
        return $output[0];
    }

    return "error=no status returned by refine_results_with_ml.pl\n";
}

###############################################################################
# my $status = refine_it($resfilepath);
#
# Call refine_results_with_ml.pl --refine. Return value is one of:
#
# not_created
# created
# creating=%d%%
# error=...
#
# Note, these all end with a newline (CRLF or LF).
###############################################################################
sub refine_it {
    my ($resfilepath) = @_;

    my $perl_path = perl_path();

    my $pipe_in;
    unless (open($pipe_in, "\"$perl_path\" ../bin/refine_results_with_ml.pl --refine --resfile ".common_subs::shellQuote($resfilepath).' |')) {
        return "error=Could not open socket to refine_results_with_ml.pl\n";
    }

    my @output = <$pipe_in>;

    if (@output) {
        return $output[0];
    }

    return "error=no status returned by refine_results_with_ml.pl\n";
}

###############################################################################
# my $resfilepath = fetch_result_file_name($taskID, $sessionID_parameter);
#
# Call ms-searchcontrol.exe --result_file_name. If it succeeds, return the
# filepath associated with the task. If anything fails, or if the task has
# no associated filepath, return undef.
###############################################################################
sub fetch_result_file_name {
    my ($taskID, $sessionID_parameter) = @_;

    my $resfilepath;

    if (open(my $pipe, '"../bin/ms-searchcontrol.exe" --result_file_name --taskID '.common_subs::shellQuote($taskID).' '.$sessionID_parameter.' |')) {
      while (<$pipe>) {
        if (/^filename=(.+)/i) {
          chomp($resfilepath = $1);
          last;
        }
      }
      1 while (<$pipe>); # flush the pipe so that zombies won't arise
    } else {
      # ignore error
    }

    return $resfilepath;
}

###############################################################################
# remove_flag_file_if_exists($resfilepath);
#
# Remove the flag file (.flg) if it exists. Also remove the batch log file
# if it exists.
###############################################################################
sub remove_flag_file_if_exists {
    my ($resfilepath) = @_;
    
    if (defined $resfilepath) {
      my $flg = $resfilepath . '.flg';
      if (-e $flg) {
        unlink $flg; # ignore error
      }
      my $log = $resfilepath . '.log'; # batch log file
      if (-e $log) {
        unlink $log; # ignore error
      }
    }
}
